#coding:utf-8
import sys, os, time
import collections
import sens, contr, kbh
""" Файл nna_06.py
    
    Скрипт для автоматизации процесса ректификации с критерием отбора основной
    фракции (тела) по температуре в нижней части колонны. Скорость отбора тела
    регулируется так, чтобы эта температура находилась вблизи уставки Tset. 
    Сама же уставка на каждом шаге главного цикла программы корректируется
    с учетом колебаний атмосферного давления.
    
    OldBean, 10.03.2017"""
#-------------------------------------------------------------------------------
# Тестируем датчики температуры (DS18B20)
#-------------------------------------------------------------------------------
hid = 0x180a353 # ID датчика температуры куба
cid = 0x18c5c68 # ID датчика температуры в нижней части колонны
did = 0x18c6520 # ID датчика температуры на входе дефлегматора
wid = 0x180c746 # ID датчика температуры выходящей воды охлаждения дефлегматора.

tids = (hid, cid, did, wid) # Вспомогательный список идентификаторов датчиков
Ts = sens.DS18B20.Ts() # Проверим наличие и работоспособность датчиков
flag = False
for id in tids:
  if Ts.get(id) != None: print "0x%012x :% 7.3f°C" % (id, Ts[id])
  else: 
    print "Датчик с ID = 0x%012x отсутствует или неисправен" % (id)
    flag = True
if flag:
  print "Один или более датчиков DS18B20 отсутствуют или неисправны. Завершение работы."
  sys.exit(1)
#-------------------------------------------------------------------------------
# Датчик RMS
#-------------------------------------------------------------------------------
vs = sens.RMS() # Создаем объект - датчик RMS
if vs != None: # Датчик в наличии и функционирует
  V = vs.V
  if V > 100: print "Действующее напряжение сети %d В" % (V)
  else:
    print "Слишком низкое напряжение сети. Завершение работы."
    sys.exit(2)
else: pass # Ну на нет и суда нет...
#-------------------------------------------------------------------------------
# Датчик атмосферного давления BMP180
#-------------------------------------------------------------------------------
ps = sens.BMP180() # Создаем объект - датчик атмосферного давления
if ps != None: # Датчик в наличии и функционирует
  print "Атмосферное давление: \t%5.1f мм.рт.ст." % (ps.P)
  print "Температура окружающей среды:\t%5.2f°C" % (ps.T)
else:
  print "Для данного алгоритма управления отбором необходим датчик атмосферного давления. Завершение работы."
  sys.exit(3)
#-------------------------------------------------------------------------------
# Создаем объекты - контроллеры ТЭНа и клапана отбора
#-------------------------------------------------------------------------------
teh = contr.TEH(vs) # Создадим объект "ТЭН"
if teh == None:
  print "Контроллер ТЭНа отсутствует или неисправен. Завершение работы."
  sys.exit(3)
sd = contr.SD() # Создадим объект "устройство отбора"
#-------------------------------------------------------------------------------
# Переведем консоль в неблокирующий режим работы
#-------------------------------------------------------------------------------
timeout = 3000 # В мс. Этим параметром можно регулировать "период" главного 
               # цикла. Сейчас (при 3000) он будет приблизительно 4 сек
tt = kbh.TT(timeout)
#-------------------------------------------------------------------------------
""" В момент активации режима регулирования отбора измеряется и запоминается
    температура в нижней части колонны Tset0 и атмосферное давление Pa0. Далее,
    на каждом шаге, вычисляется уставка с корррекцией на атмосферное лавление: 
    Tset = Tset0 + 0.033*(P - Pa0), где P - текущее атмосферное давление. 
    Пока реализован простейший ("релейный") регулятор и интегратором: если
    текущая температура меньше уставки - отбор увеличивается на dQ, если
    больше - уменьшается. Дополнительно вводится небольшая постоянная добавка
    dQp. В результате чего регулятор "старается" побыстрее добиться максимальной
    скорости отбора, при соблюдении критерия близости температуры в колонне
    к уставке"""
dQ  = 1.0
dQp = 0.1
print "dQ = %5.1f мл/час, dQp = %5.2f" % (dQ, dQp)
#-------------------------------------------------------------------------------
""" Карта режимов работы установки. Каждый режим представлен именованным 
    кортежем, содержащим следующие параметры:
      W     - мощность нагрева ТЭНа, Вт
      stab  - необходимость стабилизации мощности нагрева ТЭНа, True/False
      Q     - номинальная скорость отбора, мл/час
      ststp - необходимость регулирования скорости отбора, True/Flase
      Fmax  - максимальный суммарный расход для данного режима (-1 - нет)
   Для изменения параметров режимов, измените соотвествующие значения
   кортежей """

mrec = collections.namedtuple('mrec', 'W stab Q ststp Fmax')
modes = [] # Список режимов
modes.append(mrec(0,    False, 0,    False, -1))     # 0 - Только мониторинг
modes.append(mrec(2000, False, 0,    False, -1))     # 1 - Разгон
modes.append(mrec(600,  True,  0,    False, -1))     # 2 - Холостой ход
modes.append(mrec(600,  True,  40,   False,  120))   # 3 - Отбор голов
modes.append(mrec(600,  True,  400,  True,   1000))  # 4 - Отбор подголовников
modes.append(mrec(600,  True,  400,  True,   5000))  # 5 - Отбор тела
modes.append(mrec(600,  True,  400,  False,  1000))  # 6 - Отбор хвостов
modes.append(mrec(2000, True,  1000, False, -1))     # 7 - Пропарка колонны

mode = old_mode = 0 # В начале работы - режим 0
#-------------------------------------------------------------------------------
tst = rst = time.time() # Начало общего отсчета времени и текущего режима
#-------------------------------------------------------------------------------
""" Разбираемся с файлом журнала - продолжить старый, если есть, или начать новый, 
    если файла журнала нет или слишком мал ( < 128 байтов)"""

if os.path.exists("log") and os.path.getsize("log") > 127: 
  log = open("log", "r") # Нужно прочитать последнюю строку
  log.seek(-127, 2) # Встанeм немного недоходя до конца файла
  for line in open("log", "r"): # и считаем последнюю строку
    last_line = line
  prevStopTime = float(last_line.split()[1])
  tst -= prevStopTime*60.0 # Локальное время будет продолжаться с конца
                           # предыдущей сессии
  log.close()
  log = open("log", "a") # А теперь откроем его уже для добавления
else: # В противном случае создаем новый файл журнала
  log = open("log", "w")
#-------------------------------------------------------------------------------
ssFlag = False # Индикатор включения (True) и выключения старт-стоп
#-------------------------------------------------------------------------------
""" Подготовка закончена - входим в главный цикл приложения"""
while True:
  # Обрабатываем состояние клавиатуры...........................................
  if tt.kbhit(): # Какая-то клавиша была нажата
    ch = tt.getch()
    if ch == 'q': break # Выходим из цикла - конец работы
    elif ch == 's' and modes[mode].ststp: # Включение регулирования отбора
      Tset0 = Tset = Ts[cid] # Запоминаем текущую температуру в колонне
      Pa0 = P # Запомним текущее атмосферное давление
      ssFlag = True
    elif ch == 'r' and modes[mode].ststp: # Выключение регулирования отбора
      ssFlag = False
    elif ord(ch) >= 48 and ord(ch) <= 57: # Ввели номер режима
      old_mode = mode; mode = int(ch)
  
  # Протоколирование процесса...................................................
  """ Измеряем все доступные параметры процесса, выводим их на консоль и в файл
      журнала. """
  now = time.time(); dtm = (now - tst)/60.0; dtrm = (now - rst)/60.0
  # ss - строка для вывода на консоль
  ss = "% 7.2f % 7.2f\x1b[33m %d\x1b[0m" % (dtm, dtrm, mode)
  # s - строка для записи информации в журнал
  s = time.strftime("%y.%m.%d.%H.%M.%S")
  s += "% 7.2f % 7.2f %d" % (dtm, dtrm, mode)

  Ts = sens.DS18B20.Ts()
  for id in tids: # Текущие температуры
    if id == cid: ss += "\x1b[32m% 7.3f\x1b[0m" % (Ts[id])
    elif id == wid: ss += "% 6.1f" % (Ts[id])
    else: ss += "% 6.2f" % (Ts[id])
    s +=  "% 7.3f" % (Ts[id])

  cW = teh.W # Текущая мощность нагрева
  ss += "\x1b[31m %4d\x1b[0m" % (cW); s +=  " %4d" % (cW)
#  if vs != None: # Если есть датчик RMS - выводим напряжение сети
#    V = vs.V; ss += " %3d" % (V); s +=  " %3d" % (V)

  if ps != None: # Если есть датчик атмосферного давления - выводим давление
    P = ps.P; ss += "% 5.1f" % (P); s += "% 5.1f" % (P)
  else: P = 760.0 # Просто "затычка" для отладки

  if sd != None: # Если есть клапан отбора - выводим скорость отбора и расход
    Q = sd.Q;
    if ssFlag: # Режим старт-стопа включен - подсветим значение отбора для индикации
      ss += "\x1b[36m %4d\x1b[0m" % (Q)
    else: ss += " %4d" % (Q);
    s += " %4d" % (Q)

    F = sd.F; ss += " %5d" % (F); s += " %5d" % (F)

  if modes[mode].ststp and ssFlag: # Для отладки. Регулирование активировано
    ss += "\x1b[35m %7.3f\x1b[0m" % (Tset); s += " %7.3f" % (Tset)

  log.write(s + "\n"); log.flush()
  print ss
  
  # Переключаем режимы в зависимости от условий.................................
  if mode == 1:  # Режим "разгона"
    if Ts[cid] >= 60: # Куб закипел - пора сбрасывать мощность  
      old__mode = mode; mode = 2
  # Если не было смены режима, но есть ограничение на общий суммарный расход
  if old_mode == mode and modes[mode].Fmax > 0: 
    if F >= modes[mode].Fmax: # Завершение по достижению заданного расхода
      old__mode = mode; mode = 2

  if Ts[wid] >= 40: # Если охлаждение дефлегматора недостаточно для работы
    teh.W = 0; sd.Q = 0 # Отключаем нагрев и отбор
    old__mode = mode; mode = 0 # Переходим в режим мониторинга

  if mode == 6: # Отбор хвостов (для оборотного спирта) прекращаем отбор,
    if Ts[did] >= 80: # если начался рост температуры в дефлегматоре
      old__mode = mode; mode = 2

  # Отрабатываем смену режима...................................................
  if old_mode != mode:
    """Для режимов без стабилизации мощности (разгон, пропарка клонны и т.п.)
       мощность устанавливается только один раз при смене режима"""
    if not modes[mode].stab and mode != 0: teh.W = modes[mode].W               
    sd.Q = modes[mode].Q # Это же и для скорости отбора
    sd.FZ() # Сброс счетчика расхода при смене режима
    if modes[mode].ststp: # Но, если предусмотрено регурирование отбора,
      curQ = modes[mode].Q # то будем использовать текущую скорость отбора
      ssFlag = False # В самом начале режима регулировка дезактивирована
    if modes[old_mode].ststp: ssFlag = False # Почистим флаг на всякий случай :)
    rst = now # Фиксируем момент начала текущего режима
    old_mode = mode # Ну теперь вроде бы все сделали для смены режима

  # Стабилизация мощности нагрева, если предусмотрено режимом...................
  """А для режимов со стабилизацией мощности, мощность обновляется на каждом
     такте главного цикла приложения (естественно, с учетом напряжения сети)"""
  if modes[mode].stab: # Предусмотрена стабилизацияя мощности нагрева ТЭНа
    teh.W = modes[mode].W

  # Регулирование скорости отбора, если предусмотрено режимом...................
  """ Если текущий режим предусматривает регулирование скорости отбора
      (параметр modes[].ststp равен True) и если сам регулятор активирован 
      (флаг ssFlag равен True, активация - нажатие клавиши 's', дезактивация -
      нажатием 'r'), то выполняются следующие действия:
      1. Корректируется уставка с учетом колебаний атмосферного давления
      2. В зависимости от температуры в нижней части колонны (относительно 
         уставки), скорость отбора спирта либо увеличивается, либо уменьшается,
         либо остается неизменной (релейный алгоритм с интегратором)"""
  if modes[mode].ststp and ssFlag: # Регулирование активировано
    Tset = Tset0 + 0.033*(P - Pa0) # Корректируем уставку
    if Ts[cid] > Tset + 0.0625: # 0.0625 - квант температурного датчика DS18B20
      curQ -= dQ
      if curQ < 0: curQ = 0     # Ограничитель на отбор снизу
    elif Ts[cid] < Tset - 0.0625:
      curQ += dQ
      if curQ > 800: curQ = 800 # Ограничитель на отбор сверху
    """ Помимо вышеприведенного регулятора введем медленное (независимое)
        увеличение скорости отбора с малым шагом dQp т.к. нам желательно
        чтобы скорость отбора достигала максимального значения (естественно, 
        с учетом предыдущего критерия """
    curQ += dQp
    sd.Q = curQ # Модифицируем текущую скорость отбора в контроллере
